@@ -9,23 +9,46 @@
 #include "rimeengine.h"
 #include "rimestate.h"
 #include <fcitx/candidatelist.h>
+#include <limits>
+#include <memory>
+#include <vector>
 
 namespace fcitx {
 
 class RimeCandidateWord : public CandidateWord {
 public:
     RimeCandidateWord(RimeEngine *engine, const RimeCandidate &candidate,
-                      KeySym sym);
+                      KeySym sym, int idx);
 
     void select(InputContext *inputContext) const override;
 
 private:
     RimeEngine *engine_;
     KeySym sym_;
+    int idx_;
 };
 
+#ifndef FCITX_RIME_NO_SELECT_CANDIDATE
+class RimeGlobalCandidateWord : public CandidateWord {
+public:
+    RimeGlobalCandidateWord(RimeEngine *engine, const RimeCandidate &candidate,
+                            int idx);
+
+    void select(InputContext *inputContext) const override;
+
+private:
+    RimeEngine *engine_;
+    int idx_;
+};
+#endif
+
 class RimeCandidateList final : public CandidateList,
-                                public PageableCandidateList {
+                                public PageableCandidateList
+#ifndef FCITX_RIME_NO_SELECT_CANDIDATE
+    ,
+                                public BulkCandidateList
+#endif
+{
 public:
     RimeCandidateList(RimeEngine *engine, InputContext *ic,
                       const RimeContext &context);
@@ -62,6 +85,11 @@ public:
 
     bool usedNextBefore() const override { return true; }
 
+#ifndef FCITX_RIME_NO_SELECT_CANDIDATE
+    const CandidateWord &candidateFromAll(int idx) const override;
+    int totalSize() const override;
+#endif
+
 private:
     void checkIndex(int idx) const {
         if (idx < 0 && idx >= size()) {
@@ -76,7 +104,12 @@ private:
     bool hasNext_ = false;
     CandidateLayoutHint layout_ = CandidateLayoutHint::NotSet;
     int cursor_ = -1;
+
     std::vector<std::unique_ptr<CandidateWord>> candidateWords_;
+
+    mutable size_t maxSize_ = std::numeric_limits<size_t>::max();
+    mutable std::vector<std::unique_ptr<RimeGlobalCandidateWord>>
+        globalCandidateWords_;
 };
 } // namespace fcitx
 
