@@ -6,9 +6,9 @@
 #ifndef _FCITX_RIMEENGINE_H_
 #define _FCITX_RIMEENGINE_H_
 
-#include "rimeservice.h"
 #include "rimesession.h"
 #include "rimestate.h"
+#include <cstdint>
 #include <fcitx-config/configuration.h>
 #include <fcitx-config/iniparser.h>
 #include <fcitx-utils/event.h>
@@ -27,19 +27,38 @@
 #include <fcitx/menu.h>
 #include <memory>
 #include <rime_api.h>
+#include <string>
+#include <unordered_map>
+
+#ifndef FCITX_RIME_NO_DBUS
+#include "rimeservice.h"
+#endif
 
 namespace fcitx {
 
 class RimeState;
 
+enum class SharedStatePolicy { FollowGlobalConfig, All, Program, No };
+
+FCITX_CONFIG_ENUM_NAME_WITH_I18N(SharedStatePolicy,
+                                 N_("Follow Global Configuration"), N_("All"),
+                                 N_("Program"), N_("No"));
+
+enum class PreeditMode { No, ComposingText, CommitPreview };
+
+FCITX_CONFIG_ENUM_NAME_WITH_I18N(PreeditMode, N_("Do not show"),
+                                 N_("Composing text"), N_("Commit preview"))
+
 FCITX_CONFIGURATION(
     RimeEngineConfig,
-    Option<bool> showPreeditInApplication{this, "PreeditInApplication",
-                                          _("Show preedit within application"),
-                                          true};
+    OptionWithAnnotation<PreeditMode, PreeditModeI18NAnnotation> preeditMode{
+        this, "PreeditMode", _("Preedit Mode"),
+        isAndroid() ? PreeditMode::No : PreeditMode::ComposingText};
+    Option<SharedStatePolicy> sharedStatePolicy{
+        this, "InputState", _("Shared Input State"), SharedStatePolicy::All};
     Option<bool> preeditCursorPositionAtBeginning{
         this, "PreeditCursorPositionAtBeginning",
-        _("Fix embedded preedit cursor at the beginning of the preedit"), true};
+        _("Fix embedded preedit cursor at the beginning of the preedit"), !isAndroid()};
     Option<bool> commitWhenDeactivate{
         this, "Commit when deactivate",
         _("Commit current text when deactivating"), true};
@@ -64,6 +83,12 @@ FCITX_CONFIGURATION(
 #endif
 );
 
+class RimeOptionAction : public Action {
+public:
+    // This is used to save the option when we need to release the session.
+    virtual std::optional<std::string> snapshotOption(InputContext *ic) = 0;
+};
+
 class RimeEngine final : public InputMethodEngineV2 {
 public:
     RimeEngine(Instance *instance);
@@ -80,10 +105,6 @@ public:
     void save() override;
     auto &factory() { return factory_; }
 
-    void updateAction(InputContext *inputContext) {
-        imAction_->update(inputContext);
-    }
-
     const Configuration *getConfig() const override { return &config_; }
     void setConfig(const RawConfig &config) override {
         config_.load(config, true);
@@ -108,7 +129,13 @@ public:
     RimeState *state(InputContext *ic);
     RimeSessionPool &sessionPool() { return sessionPool_; }
 
+#ifndef FCITX_RIME_NO_DBUS
     FCITX_ADDON_DEPENDENCY_LOADER(dbus, instance_->addonManager());
+#endif
+
+    void blockNotificationFor(uint64_t usec);
+    const auto &schemas() const { return schemas_; }
+    const auto &optionActions() const { return optionActions_; };
 
 private:
     static void rimeNotificationHandler(void *context_object,
@@ -119,19 +146,30 @@ private:
     void deploy();
     void sync();
     void updateSchemaMenu();
-    void notify(const std::string &type, const std::string &value);
-    void releaseAllSession();
+    void updateActionsForSchema(const std::string &schema);
+    void notify(RimeSessionId session, const std::string &type,
+                const std::string &value);
+    void releaseAllSession(const bool snapshot = false);
     void updateAppOptions();
+    void refreshStatusArea(InputContext &ic);
+    void refreshStatusArea(RimeSessionId session);
+    void updateStatusArea(RimeSessionId session);
+    void refreshSessionPoolPolicy();
+    PropertyPropagatePolicy getSharedStatePolicy();
 
+    bool constructed_ = false;
+    std::string sharedDataDir_;
     IconTheme theme_;
     Instance *instance_;
     EventDispatcher eventDispatcher_;
     rime_api_t *api_;
-    bool firstRun_ = true;
+    static bool firstRun_;
     uint64_t blockNotificationBefore_ = 0;
+    uint64_t lastKeyEventTime_ = 0;
     FactoryFor<RimeState> factory_;
 
     std::unique_ptr<Action> imAction_;
+    SimpleAction separatorAction_;
     SimpleAction deployAction_;
     SimpleAction syncAction_;
 
@@ -141,15 +179,20 @@ private:
 
     FCITX_ADDON_DEPENDENCY_LOADER(notifications, instance_->addonManager());
 
+    std::unordered_set<std::string> schemas_;
     std::list<SimpleAction> schemActions_;
+    std::unordered_map<std::string,
+                       std::list<std::unique_ptr<RimeOptionAction>>>
+        optionActions_;
     Menu schemaMenu_;
 #ifdef FCITX_RIME_LOAD_PLUGIN
     std::unordered_map<std::string, Library> pluginPool_;
 #endif
-    std::unique_ptr<EventSourceTime> timeEvent_;
     std::unique_ptr<HandlerTableEntry<EventHandler>> globalConfigReloadHandle_;
 
+#ifndef FCITX_RIME_NO_DBUS
     RimeService service_{this};
+#endif
     RimeSessionPool sessionPool_;
 };
 
