@@ -43,9 +43,11 @@ class RimeSessionPool {
 public:
     RimeSessionPool(RimeEngine *engine, PropertyPropagatePolicy initialPolicy);
 
+    PropertyPropagatePolicy propertyPropagatePolicy() const { return policy_; }
     void setPropertyPropagatePolicy(PropertyPropagatePolicy policy);
 
-    std::shared_ptr<RimeSessionHolder> requestSession(InputContext *ic);
+    std::tuple<std::shared_ptr<RimeSessionHolder>, bool>
+    requestSession(InputContext *ic);
 
     RimeEngine *engine() const { return engine_; }
 
