@@ -6,6 +6,9 @@
 
 #include "rimestate.h"
 #include "rimecandidate.h"
+#include "rimeengine.h"
+#include <fcitx-utils/stringutils.h>
+#include <fcitx-utils/textformatflags.h>
 #include <fcitx-utils/utf8.h>
 #include <fcitx/candidatelist.h>
 #include <fcitx/inputcontext.h>
@@ -30,13 +33,22 @@ RimeState::RimeState(RimeEngine *engine, InputContext &ic)
 
 RimeState::~RimeState() {}
 
-RimeSessionId RimeState::session() {
-    if (!session_) {
-        session_ = engine_->sessionPool().requestSession(&ic_);
+RimeSessionId RimeState::session(bool requestNewSession) {
+    if (!session_ && requestNewSession) {
+        auto [sessionHolder, isNewSession] =
+            engine_->sessionPool().requestSession(&ic_);
+        session_ = sessionHolder;
+        if (isNewSession) {
+            restore();
+        } else {
+            savedCurrentSchema_.clear();
+            savedOptions_.clear();
+        }
     }
     if (!session_) {
         return 0;
     }
+
     return session_->id();
 }
 
@@ -82,24 +94,25 @@ std::string RimeState::subModeLabel() {
 
 void RimeState::setLatinMode(bool latin) {
     auto api = engine_->api();
-    if (!api || api->is_maintenance_mode()) {
+    if (api->is_maintenance_mode()) {
         return;
     }
-    api->set_option(session(), "ascii_mode", latin);
+    api->set_option(session(), RIME_ASCII_MODE, latin);
 }
 
 void RimeState::selectSchema(const std::string &schema) {
     auto api = engine_->api();
-    if (!api || api->is_maintenance_mode()) {
+    if (api->is_maintenance_mode()) {
         return;
     }
-    api->set_option(session(), "ascii_mode", false);
+    engine_->blockNotificationFor(30000);
+    api->set_option(session(), RIME_ASCII_MODE, false);
     api->select_schema(session(), schema.data());
 }
 
 void RimeState::keyEvent(KeyEvent &event) {
     auto api = engine_->api();
-    if (!api || api->is_maintenance_mode()) {
+    if (api->is_maintenance_mode()) {
         return;
     }
     auto session = this->session();
@@ -136,12 +149,34 @@ void RimeState::keyEvent(KeyEvent &event) {
     }
 }
 
+#ifndef FCITX_RIME_NO_SELECT_CANDIDATE
+void RimeState::selectCandidate(InputContext *inputContext, int idx,
+                                bool global) {
+    auto api = engine_->api();
+    if (api->is_maintenance_mode()) {
+        return;
+    }
+    auto session = this->session();
+    if (!session) {
+        return;
+    }
+    if (global) {
+        api->select_candidate(session, idx);
+    } else {
+        api->select_candidate_on_current_page(session, idx);
+    }
+    RIME_STRUCT(RimeCommit, commit);
+    if (api->get_commit(session, &commit)) {
+        inputContext->commitString(commit.text);
+        api->free_commit(&commit);
+    }
+    updateUI(inputContext, false);
+}
+#endif
+
 bool RimeState::getStatus(
     const std::function<void(const RimeStatus &)> &callback) {
     auto api = engine_->api();
-    if (!api) {
-        return false;
-    }
     auto session = this->session();
     if (!session) {
         return false;
@@ -155,15 +190,9 @@ bool RimeState::getStatus(
     return true;
 }
 
-void RimeState::updatePreedit(InputContext *ic, const RimeContext &context) {
+Text preeditFromRimeContext(const RimeContext &context, TextFormatFlags flag) {
     Text preedit;
-    Text clientPreedit;
 
-    TextFormatFlags flag;
-    if (engine_->config().showPreeditInApplication.value() &&
-        ic->capabilityFlags().test(CapabilityFlag::Preedit)) {
-        flag = TextFormatFlag::Underline;
-    }
     do {
         if (context.composition.length == 0) {
             break;
@@ -181,11 +210,6 @@ void RimeState::updatePreedit(InputContext *ic, const RimeContext &context) {
             preedit.append(std::string(context.composition.preedit,
                                        context.composition.sel_start),
                            flag);
-            if (context.commit_text_preview) {
-                clientPreedit.append(std::string(context.commit_text_preview,
-                                                 context.composition.sel_start),
-                                     flag);
-            }
         }
 
         /* converting candidate */
@@ -195,12 +219,6 @@ void RimeState::updatePreedit(InputContext *ic, const RimeContext &context) {
                     &context.composition.preedit[context.composition.sel_start],
                     &context.composition.preedit[context.composition.sel_end]),
                 flag | TextFormatFlag::HighLight);
-            if (context.commit_text_preview) {
-                clientPreedit.append(
-                    std::string(&context.commit_text_preview[context.composition
-                                                                 .sel_start]),
-                    flag | TextFormatFlag::HighLight);
-            }
         }
 
         /* remaining input to convert */
@@ -215,16 +233,43 @@ void RimeState::updatePreedit(InputContext *ic, const RimeContext &context) {
         preedit.setCursor(context.composition.cursor_pos);
     } while (0);
 
-    if (engine_->config().showPreeditInApplication.value() &&
-        ic->capabilityFlags().test(CapabilityFlag::Preedit)) {
-        clientPreedit = preedit;
-    } else {
-        ic->inputPanel().setPreedit(preedit);
+    return preedit;
+}
+
+void RimeState::updatePreedit(InputContext *ic, const RimeContext &context) {
+    PreeditMode mode = ic->capabilityFlags().test(CapabilityFlag::Preedit)
+                           ? *engine_->config().preeditMode
+                           : PreeditMode::No;
+
+    switch (mode) {
+    case PreeditMode::No:
+        ic->inputPanel().setPreedit(
+            preeditFromRimeContext(context, TextFormatFlag::NoFlag));
+        break;
+    case PreeditMode::CommitPreview: {
+        ic->inputPanel().setPreedit(
+            preeditFromRimeContext(context, TextFormatFlag::NoFlag));
+        if (context.commit_text_preview) {
+            Text clientPreedit;
+            clientPreedit.append(context.commit_text_preview,
+                                 TextFormatFlag::Underline);
+            if (*engine_->config().preeditCursorPositionAtBeginning) {
+                clientPreedit.setCursor(0);
+            } else {
+                clientPreedit.setCursor(clientPreedit.textLength());
+            }
+            ic->inputPanel().setClientPreedit(clientPreedit);
+        }
+    } break;
+    case PreeditMode::ComposingText: {
+        Text clientPreedit =
+            preeditFromRimeContext(context, TextFormatFlag::Underline);
+        if (*engine_->config().preeditCursorPositionAtBeginning) {
+            clientPreedit.setCursor(0);
+        }
+        ic->inputPanel().setClientPreedit(clientPreedit);
+    } break;
     }
-    if (engine_->config().preeditCursorPositionAtBeginning.value()) {
-        clientPreedit.setCursor(0);
-    }
-    ic->inputPanel().setClientPreedit(clientPreedit);
 }
 
 void RimeState::updateUI(InputContext *ic, bool keyRelease) {
@@ -233,11 +278,10 @@ void RimeState::updateUI(InputContext *ic, bool keyRelease) {
         inputPanel.reset();
     }
     bool oldEmptyExceptAux = emptyExceptAux(inputPanel);
-    engine_->updateAction(ic);
 
     do {
         auto api = engine_->api();
-        if (!api || api->is_maintenance_mode()) {
+        if (api->is_maintenance_mode()) {
             return;
         }
         auto session = this->session();
@@ -273,6 +317,7 @@ void RimeState::updateUI(InputContext *ic, bool keyRelease) {
         inputPanel.setAuxDown(Text());
     }
     if (newEmptyExceptAux && lastMode_ != subMode()) {
+        engine_->blockNotificationFor(30000);
         engine_->instance()->showInputMethodInformation(ic);
         ic->updateUserInterface(UserInterfaceComponent::StatusArea);
     }
@@ -297,4 +342,49 @@ void RimeState::commitPreedit(InputContext *ic) {
         api->free_context(&context);
     }
 }
+
+void RimeState::snapshot() {
+    if (!session(false)) {
+        return;
+    }
+    getStatus([this](const RimeStatus &status) {
+        if (!status.schema_id) {
+            return;
+        }
+        savedCurrentSchema_ = status.schema_id;
+        savedOptions_.clear();
+        if (savedCurrentSchema_.empty()) {
+            return;
+        }
+        auto &optionActions = engine_->optionActions();
+        auto iter = optionActions.find(savedCurrentSchema_);
+        if (iter == optionActions.end()) {
+            return;
+        }
+        for (const auto &option : iter->second) {
+            if (auto savedOption = option->snapshotOption(&ic_)) {
+                savedOptions_.push_back(std::move(*savedOption));
+            }
+        }
+    });
+}
+
+void RimeState::restore() {
+    if (savedCurrentSchema_.empty()) {
+        return;
+    }
+    if (!engine_->schemas().count(savedCurrentSchema_)) {
+        return;
+    }
+
+    selectSchema(savedCurrentSchema_);
+    for (const auto &option : savedOptions_) {
+        if (stringutils::startsWith(option, "!")) {
+            engine_->api()->set_option(session(), option.c_str() + 1, false);
+        } else {
+            engine_->api()->set_option(session(), option.c_str(), true);
+        }
+    }
+}
+
 } // namespace fcitx
