@@ -13,6 +13,8 @@
 #include <memory>
 #include <rime_api.h>
 
+#define RIME_ASCII_MODE "ascii_mode"
+
 namespace fcitx {
 
 class RimeEngine;
@@ -25,6 +27,9 @@ public:
 
     void clear();
     void keyEvent(KeyEvent &event);
+#ifndef FCITX_RIME_NO_SELECT_CANDIDATE
+    void selectCandidate(InputContext *inputContext, int idx, bool global);
+#endif
     bool getStatus(const std::function<void(const RimeStatus &)> &);
     void updatePreedit(InputContext *ic, const RimeContext &context);
     void updateUI(InputContext *ic, bool keyRelease);
@@ -34,13 +39,19 @@ public:
     std::string subModeLabel();
     void setLatinMode(bool latin);
     void selectSchema(const std::string &schemaId);
-    RimeSessionId session();
+    RimeSessionId session(bool requestNewSession = true);
+
+    void snapshot();
+    void restore();
 
 private:
     std::string lastMode_;
     RimeEngine *engine_;
     InputContext &ic_;
     std::shared_ptr<RimeSessionHolder> session_;
+
+    std::string savedCurrentSchema_;
+    std::vector<std::string> savedOptions_;
 };
 } // namespace fcitx
 
